import os
import ssl
import shutil
import pytest
import asyncio
import logging
from pathlib import Path
from contextlib import suppress
from omegaconf import OmegaConf
from pytest_httpserver import HTTPServer

from bbot.core import CORE
from bbot.core.helpers.misc import execute_sync_or_async
from bbot.core.helpers.interactsh import server_list as interactsh_servers


test_config = OmegaConf.load(Path(__file__).parent / "test.conf")
if test_config.get("debug", False):
    os.environ["BBOT_DEBUG"] = "True"
    logging.getLogger("bbot").setLevel(logging.DEBUG)
    CORE.logger.log_level = logging.DEBUG
else:
    # silence stdout + trace
    root_logger = logging.getLogger()
    for h in root_logger.handlers:
        h.addFilter(lambda x: x.levelname not in ("STDOUT", "TRACE"))

CORE.merge_default(test_config)


@pytest.fixture
def assert_all_responses_were_requested() -> bool:
    return False


@pytest.fixture
def bbot_httpserver():
    server = HTTPServer(host="127.0.0.1", port=8888, threaded=True)
    server.start()

    yield server

    server.clear()
    if server.is_running():
        server.stop()

    # this is to check if the client has made any request where no
    # `assert_request` was called on it from the test

    server.check_assertions()
    server.clear()


@pytest.fixture
def bbot_httpserver_ssl():
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    current_dir = Path(__file__).parent
    keyfile = str(current_dir / "testsslkey.pem")
    certfile = str(current_dir / "testsslcert.pem")
    context.load_cert_chain(certfile, keyfile)
    server = HTTPServer(host="127.0.0.1", port=9999, ssl_context=context, threaded=True)
    server.start()

    yield server

    server.clear()
    if server.is_running():
        server.stop()

    # this is to check if the client has made any request where no
    # `assert_request` was called on it from the test

    server.check_assertions()
    server.clear()


@pytest.fixture
def non_mocked_hosts() -> list:
    return ["127.0.0.1", "localhost", "raw.githubusercontent.com"] + interactsh_servers


@pytest.fixture
def bbot_httpserver_allinterfaces():
    server = HTTPServer(host="0.0.0.0", port=5556, threaded=True)
    server.start()

    yield server

    server.clear()
    if server.is_running():
        server.stop()
    server.check_assertions()
    server.clear()


class Interactsh_mock:
    def __init__(self, name):
        self.name = name
        self.log = logging.getLogger(f"bbot.interactsh.{self.name}")
        self.interactions = []
        self.correlation_id = "deadbeef-dead-beef-dead-beefdeadbeef"
        self.stop = False
        self.poll_task = None

    def mock_interaction(self, subdomain_tag, msg=None):
        self.log.info(f"Mocking interaction to subdomain tag: {subdomain_tag}")
        if msg is not None:
            self.log.info(msg)
        self.interactions.append(subdomain_tag)

    async def register(self, callback=None):
        if callable(callback):
            self.poll_task = asyncio.create_task(self.poll_loop(callback))
        return "fakedomain.fakeinteractsh.com"

    async def deregister(self, callback=None):
        self.stop = True
        if self.poll_task is not None:
            self.poll_task.cancel()
            with suppress(BaseException):
                await self.poll_task

    async def poll_loop(self, callback=None):
        while not self.stop:
            data_list = await self.poll(callback)
            if not data_list:
                await asyncio.sleep(1)
                continue

    async def poll(self, callback=None):
        poll_results = []
        for subdomain_tag in self.interactions:
            result = {"full-id": f"{subdomain_tag}.fakedomain.fakeinteractsh.com", "protocol": "HTTP"}
            poll_results.append(result)
            if callback is not None:
                await execute_sync_or_async(callback, result)
        self.interactions = []
        return poll_results


import threading
import http.server
import socketserver
import urllib.request


class Proxy(http.server.SimpleHTTPRequestHandler):
    protocol_version = "HTTP/1.0"
    server_version = "Proxy"
    urls = []

    def do_GET(self):
        self.urls.append(self.path)

        # Extract host and port from path
        netloc = urllib.parse.urlparse(self.path).netloc
        host, _, port = netloc.partition(":")

        # Fetch the content
        conn = http.client.HTTPConnection(host, port if port else 80)
        conn.request("GET", self.path, headers=self.headers)
        response = conn.getresponse()

        # Send the response back to the client
        self.send_response(response.status)
        for header, value in response.getheaders():
            self.send_header(header, value)
        self.end_headers()
        self.copyfile(response, self.wfile)

        response.close()
        conn.close()


@pytest.fixture
def proxy_server():
    # Set up an HTTP server that acts as a simple proxy.
    server = socketserver.ThreadingTCPServer(("localhost", 0), Proxy)

    # Start the server in a new thread.
    server_thread = threading.Thread(target=server.serve_forever, daemon=True)
    server_thread.start()

    yield server

    # Stop the server.
    server.shutdown()
    server_thread.join()


def pytest_terminal_summary(terminalreporter, exitstatus, config):  # pragma: no cover
    RED = "\033[1;31m"
    GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    BLUE = "\033[1;34m"
    CYAN = "\033[1;36m"
    RESET = "\033[0m"
    stats = terminalreporter.stats
    total_tests = len(terminalreporter._session.items)
    passed = len(stats.get("passed", []))
    skipped = len(stats.get("skipped", []))
    errors = len(stats.get("error", []))
    failed = stats.get("failed", [])

    print("\nTest Session Summary:")
    print(f"Total tests run: {total_tests}")
    print(
        f"{GREEN}Passed: {passed}{RESET}, {RED}Failed: {len(failed)}{RESET}, {YELLOW}Skipped: {skipped}{RESET}, Errors: {errors}"
    )

    if failed:
        print(f"\n{RED}Detailed failed test report:{RESET}")
        for item in failed:
            test_name = item.nodeid.split("::")[-1] if "::" in item.nodeid else item.nodeid
            file_and_line = f"{item.location[0]}:{item.location[1]}"  # File path and line number
            print(f"{BLUE}Test Name: {test_name}{RESET} {CYAN}({file_and_line}){RESET}")
            print(f"{RED}Location: {item.nodeid} at {item.location[0]}:{item.location[1]}{RESET}")
            print(f"{RED}Failure details:\n{item.longreprtext}{RESET}")


# BELOW: debugging for frozen/hung tests
# import psutil
# import traceback
# import inspect


# def _print_detailed_info():  # pragma: no cover
#     """
#     Debugging pytests hanging
#     """
#     print("=== Detailed Thread and Process Information ===\n")
#     try:
#         print("=== Threads ===")
#         for thread in threading.enumerate():
#             print(f"Thread Name: {thread.name}")
#             print(f"Thread ID: {thread.ident}")
#             print(f"Is Alive: {thread.is_alive()}")
#             print(f"Daemon: {thread.daemon}")

#             if hasattr(thread, "_target"):
#                 target = thread._target
#                 if target:
#                     qualname = (
#                         f"{target.__module__}.{target.__qualname__}"
#                         if hasattr(target, "__qualname__")
#                         else str(target)
#                     )
#                     print(f"Target Function: {qualname}")

#                     if hasattr(thread, "_args"):
#                         args = thread._args
#                         kwargs = thread._kwargs if hasattr(thread, "_kwargs") else {}
#                         arg_spec = inspect.getfullargspec(target)

#                         all_args = list(args) + [f"{k}={v}" for k, v in kwargs.items()]

#                         if inspect.ismethod(target) and arg_spec.args[0] == "self":
#                             arg_spec.args.pop(0)

#                         named_args = list(zip(arg_spec.args, all_args))
#                         if arg_spec.varargs:
#                             named_args.extend((f"*{arg_spec.varargs}", arg) for arg in all_args[len(arg_spec.args) :])

#                         print("Arguments:")
#                         for name, value in named_args:
#                             print(f"  {name}: {value}")
#                 else:
#                     print("Target Function: None")
#             else:
#                 print("Target Function: Unknown")

#             print()

#         print("=== Processes ===")
#         current_process = psutil.Process()
#         for child in current_process.children(recursive=True):
#             print(f"Process ID: {child.pid}")
#             print(f"Name: {child.name()}")
#             print(f"Status: {child.status()}")
#             print(f"CPU Times: {child.cpu_times()}")
#             print(f"Memory Info: {child.memory_info()}")
#             print()

#         print("=== Current Process ===")
#         print(f"Process ID: {current_process.pid}")
#         print(f"Name: {current_process.name()}")
#         print(f"Status: {current_process.status()}")
#         print(f"CPU Times: {current_process.cpu_times()}")
#         print(f"Memory Info: {current_process.memory_info()}")
#         print()

#     except Exception as e:
#         print(f"An error occurred: {str(e)}")
#         print("Traceback:")
#         traceback.print_exc()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_sessionfinish(session, exitstatus):
    # Remove handlers from all loggers to prevent logging errors at exit
    loggers = [logging.getLogger("bbot")] + list(logging.Logger.manager.loggerDict.values())
    for logger in loggers:
        handlers = getattr(logger, "handlers", [])
        for handler in handlers:
            logger.removeHandler(handler)

    # Wipe out BBOT home dir
    shutil.rmtree("/tmp/.bbot_test", ignore_errors=True)

    yield

    # temporarily suspend stdout capture and print detailed thread info
    # capmanager = session.config.pluginmanager.get_plugin("capturemanager")
    # if capmanager:
    #     capmanager.suspend_global_capture(in_=True)

    # _print_detailed_info()

    # if capmanager:
    #     capmanager.resume_global_capture()



from ..bbot_fixtures import *

from bbot import cli


@pytest.mark.asyncio
async def test_cli_scope(monkeypatch, capsys):
    import json

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)

    # basic target without whitelist
    monkeypatch.setattr(
        "sys.argv",
        ["bbot", "-t", "one.one.one.one", "-c", "scope.report_distance=10", "dns.minimal=false", "--json"],
    )
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == True
    lines = [json.loads(l) for l in out.splitlines()]
    dns_events = [l for l in lines if l["type"] == "DNS_NAME" and l["data"] == "one.one.one.one"]
    assert dns_events
    assert all([l["scope_distance"] == 0 and "in-scope" in l["tags"] for l in dns_events])
    assert 1 == len(
        [
            l
            for l in dns_events
            if l["module"] == "TARGET"
            and l["scope_distance"] == 0
            and "in-scope" in l["tags"]
            and "target" in l["tags"]
        ]
    )
    ip_events = [l for l in lines if l["type"] == "IP_ADDRESS" and l["data"] == "1.1.1.1"]
    assert ip_events
    assert all([l["scope_distance"] == 1 and "distance-1" in l["tags"] for l in ip_events])
    ip_events = [l for l in lines if l["type"] == "IP_ADDRESS" and l["data"] == "1.0.0.1"]
    assert ip_events
    assert all([l["scope_distance"] == 1 and "distance-1" in l["tags"] for l in ip_events])

    # with whitelist
    monkeypatch.setattr(
        "sys.argv",
        [
            "bbot",
            "-t",
            "one.one.one.one",
            "-w",
            "192.168.0.1",
            "-c",
            "scope.report_distance=10",
            "dns.minimal=false",
            "dns.search_distance=2",
            "--json",
        ],
    )
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == True
    lines = [json.loads(l) for l in out.splitlines()]
    lines = [l for l in lines if l["type"] != "SCAN"]
    assert lines
    assert not any([l["scope_distance"] == 0 for l in lines])
    dns_events = [l for l in lines if l["type"] == "DNS_NAME" and l["data"] == "one.one.one.one"]
    assert dns_events
    assert all([l["scope_distance"] == 1 and "distance-1" in l["tags"] for l in dns_events])
    assert 1 == len(
        [
            l
            for l in dns_events
            if l["module"] == "TARGET"
            and l["scope_distance"] == 1
            and "distance-1" in l["tags"]
            and "target" in l["tags"]
        ]
    )
    ip_events = [l for l in lines if l["type"] == "IP_ADDRESS" and l["data"] == "1.1.1.1"]
    assert ip_events
    assert all([l["scope_distance"] == 2 and "distance-2" in l["tags"] for l in ip_events])
    ip_events = [l for l in lines if l["type"] == "IP_ADDRESS" and l["data"] == "1.0.0.1"]
    assert ip_events
    assert all([l["scope_distance"] == 2 and "distance-2" in l["tags"] for l in ip_events])


@pytest.mark.asyncio
async def test_cli_scan(monkeypatch):
    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)

    scans_home = bbot_test_dir / "scans"

    # basic scan
    monkeypatch.setattr(
        sys,
        "argv",
        ["bbot", "-y", "-t", "127.0.0.1", "www.example.com", "-n", "test_cli_scan", "-c", "dns.disable=true"],
    )
    result = await cli._main()
    assert result == True

    scan_home = scans_home / "test_cli_scan"
    assert (scan_home / "preset.yml").is_file(), "preset.yml not found"
    assert (scan_home / "wordcloud.tsv").is_file(), "wordcloud.tsv not found"
    assert (scan_home / "output.txt").is_file(), "output.txt not found"
    assert (scan_home / "output.csv").is_file(), "output.csv not found"
    assert (scan_home / "output.json").is_file(), "output.json not found"

    with open(scan_home / "preset.yml") as f:
        text = f.read()
        assert "  dns:\n    disable: true" in text

    with open(scan_home / "output.csv") as f:
        lines = f.readlines()
        assert lines[0] == "Event type,Event data,IP Address,Source Module,Scope Distance,Event Tags,Discovery Path\n"
        assert len(lines) > 1, "output.csv is not long enough"

    ip_success = False
    dns_success = False
    output_filename = scan_home / "output.txt"
    with open(output_filename) as f:
        lines = f.read().splitlines()
        for line in lines:
            if "[IP_ADDRESS]        \t127.0.0.1\tTARGET" in line:
                ip_success = True
            if "[DNS_NAME]          \twww.example.com\tTARGET" in line:
                dns_success = True
    assert ip_success and dns_success, "IP_ADDRESS and/or DNS_NAME are not present in output.txt"


@pytest.mark.asyncio
async def test_cli_args(monkeypatch, caplog, capsys, clean_default_config):
    caplog.set_level(logging.INFO)

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)

    # show version
    monkeypatch.setattr("sys.argv", ["bbot", "--version"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert len(out.splitlines()) == 1
    assert out.count(".") > 1

    # list modules
    monkeypatch.setattr("sys.argv", ["bbot", "--list-modules"])
    result = await cli._main()
    assert result == None
    out, err = capsys.readouterr()
    # internal modules
    assert "| excavate " in out
    # output modules
    assert "| csv " in out
    # scan modules
    assert "| wayback " in out

    # output dir and scan name
    output_dir = bbot_test_dir / "bbot_cli_args_output"
    scan_name = "bbot_cli_args_scan_name"
    scan_dir = output_dir / scan_name
    assert not output_dir.exists()
    monkeypatch.setattr("sys.argv", ["bbot", "-o", str(output_dir), "-n", scan_name, "-y"])
    result = await cli._main()
    assert result == True
    assert output_dir.is_dir()
    assert scan_dir.is_dir()
    assert "[SCAN]" in open(scan_dir / "output.txt").read()
    assert "[INFO]" in open(scan_dir / "scan.log").read()
    shutil.rmtree(output_dir)

    # list module options
    monkeypatch.setattr("sys.argv", ["bbot", "--list-module-options"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| modules.wayback.urls" in out
    assert "| bool" in out
    assert "| emit URLs in addition to DNS_NAMEs" in out
    assert "| False" in out
    assert "| modules.dnsbrute.wordlist" in out
    assert "| modules.robots.include_allow" in out

    # list module options by flag
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "subdomain-enum", "--list-module-options"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| modules.wayback.urls" in out
    assert "| bool" in out
    assert "| emit URLs in addition to DNS_NAMEs" in out
    assert "| False" in out
    assert "| modules.dnsbrute.wordlist" in out
    assert not "| modules.robots.include_allow" in out

    # list module options by module
    monkeypatch.setattr("sys.argv", ["bbot", "-m", "dnsbrute", "-lmo"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert out.count("modules.") == out.count("modules.dnsbrute.")
    assert not "| modules.wayback.urls" in out
    assert "| modules.dnsbrute.wordlist" in out
    assert not "| modules.robots.include_allow" in out

    # list output module options by module
    monkeypatch.setattr("sys.argv", ["bbot", "-om", "stdout", "-lmo"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert out.count("modules.") == out.count("modules.stdout.")

    # list flags
    monkeypatch.setattr("sys.argv", ["bbot", "--list-flags"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| safe " in out
    assert "| Non-intrusive, safe to run " in out
    assert "| active " in out
    assert "| passive " in out

    # list only a single flag
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "active", "--list-flags"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert not "| safe " in out
    assert "| active " in out
    assert not "| passive " in out

    # list multiple flags
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "active", "safe", "--list-flags"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| safe " in out
    assert "| active " in out
    assert not "| passive " in out

    # no args
    monkeypatch.setattr("sys.argv", ["bbot"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "Target:\n  -t TARGET [TARGET ...]" in out

    # list modules
    monkeypatch.setattr("sys.argv", ["bbot", "-l"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| dnsbrute " in out
    assert "| httpx " in out
    assert "| robots " in out

    # list modules by flag
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "subdomain-enum", "-l"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| dnsbrute " in out
    assert "| httpx " in out
    assert not "| robots " in out

    # list modules by flag + required flag
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "subdomain-enum", "-rf", "passive", "-l"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| chaos " in out
    assert not "| httpx " in out

    # list modules by flag + excluded flag
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "subdomain-enum", "-ef", "active", "-l"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert "| chaos " in out
    assert not "| httpx " in out

    # list modules by flag + excluded module
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "subdomain-enum", "-em", "dnsbrute", "-l"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == None
    assert not "| dnsbrute " in out
    assert "| httpx " in out

    # output modules override
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-om", "csv,json", "-y"])
    result = await cli._main()
    assert result == True
    assert "Loaded 2/2 output modules, (csv,json)" in caplog.text
    caplog.clear()
    monkeypatch.setattr("sys.argv", ["bbot", "-em", "csv,json", "-y"])
    result = await cli._main()
    assert result == True
    assert "Loaded 3/3 output modules, (python,stdout,txt)" in caplog.text

    # output modules override
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-om", "subdomains", "-y"])
    result = await cli._main()
    assert result == True
    assert "Loaded 6/6 output modules, (csv,json,python,stdout,subdomains,txt)" in caplog.text

    # internal modules override
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-y"])
    result = await cli._main()
    assert result == True
    assert "Loaded 5/5 internal modules (aggregate,cloudcheck,dnsresolve,excavate,speculate)" in caplog.text
    caplog.clear()
    monkeypatch.setattr("sys.argv", ["bbot", "-em", "excavate", "speculate", "-y"])
    result = await cli._main()
    assert result == True
    assert "Loaded 3/3 internal modules (aggregate,cloudcheck,dnsresolve)" in caplog.text
    caplog.clear()
    monkeypatch.setattr("sys.argv", ["bbot", "-c", "speculate=false", "-y"])
    result = await cli._main()
    assert result == True
    assert "Loaded 4/4 internal modules (aggregate,cloudcheck,dnsresolve,excavate)" in caplog.text

    # custom target type
    out, err = capsys.readouterr()
    monkeypatch.setattr("sys.argv", ["bbot", "-t", "ORG:evilcorp", "-y"])
    result = await cli._main()
    out, err = capsys.readouterr()
    assert result == True
    assert "[ORG_STUB]          	evilcorp	TARGET" in out

    # activate modules by flag
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "passive"])
    result = await cli._main()
    assert result == True

    # unconsoleable output module
    monkeypatch.setattr("sys.argv", ["bbot", "-om", "web_report"])
    result = await cli._main()
    assert result == True

    # unresolved dependency
    monkeypatch.setattr("sys.argv", ["bbot", "-m", "wappalyzer"])
    result = await cli._main()
    assert result == True

    # enable and exclude the same module
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-m", "ffuf_shortnames", "-em", "ffuf_shortnames"])
    result = await cli._main()
    assert result == None
    assert 'Unable to add scan module "ffuf_shortnames" because the module has been excluded' in caplog.text

    # require flags
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "active", "-rf", "passive"])
    result = await cli._main()
    assert result == True

    # excluded flags
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "active", "-ef", "active"])
    result = await cli._main()
    assert result == True

    # slow modules
    monkeypatch.setattr("sys.argv", ["bbot", "-m", "bucket_digitalocean"])
    result = await cli._main()
    assert result == True

    # deadly modules
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-m", "nuclei"])
    result = await cli._main()
    assert result == False, "-m nuclei ran without --allow-deadly"
    assert "Please specify --allow-deadly to continue" in caplog.text

    # --allow-deadly
    monkeypatch.setattr("sys.argv", ["bbot", "-m", "nuclei", "--allow-deadly"])
    result = await cli._main()
    assert result == True, "-m nuclei failed to run with --allow-deadly"

    # install all deps
    monkeypatch.setattr("sys.argv", ["bbot", "--install-all-deps"])
    success = await cli._main()
    assert success == True, "--install-all-deps failed for at least one module"


@pytest.mark.asyncio
async def test_cli_customheaders(monkeypatch, caplog, capsys):
    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)
    import yaml

    # test custom headers
    monkeypatch.setattr(
        "sys.argv", ["bbot", "--custom-headers", "foo=bar", "foo2=bar2", "foo3=bar=3", "--current-preset"]
    )
    success = await cli._main()
    assert success == None, "setting custom headers on command line failed"
    captured = capsys.readouterr()
    stdout_preset = yaml.safe_load(captured.out)
    assert stdout_preset["config"]["web"]["http_headers"] == {"foo": "bar", "foo2": "bar2", "foo3": "bar=3"}

    # test custom headers invalid (no "=")
    monkeypatch.setattr("sys.argv", ["bbot", "--custom-headers", "justastring", "--current-preset"])
    result = await cli._main()
    assert result == None
    assert "Custom headers not formatted correctly (missing '=')" in caplog.text
    caplog.clear()

    # test custom headers invalid (missing key)
    monkeypatch.setattr("sys.argv", ["bbot", "--custom-headers", "=nokey", "--current-preset"])
    result = await cli._main()
    assert result == None
    assert "Custom headers not formatted correctly (missing header name or value)" in caplog.text
    caplog.clear()

    # test custom headers invalid (missing value)
    monkeypatch.setattr("sys.argv", ["bbot", "--custom-headers", "missingvalue=", "--current-preset"])
    result = await cli._main()
    assert result == None
    assert "Custom headers not formatted correctly (missing header name or value)" in caplog.text


def test_cli_config_validation(monkeypatch, caplog):
    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)

    # incorrect module option
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-c", "modules.ipnegibhor.num_bits=4"])
    cli.main()
    assert 'Could not find config option "modules.ipnegibhor.num_bits"' in caplog.text
    assert 'Did you mean "modules.ipneighbor.num_bits"?' in caplog.text

    # incorrect global option
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-c", "web_spier_distance=4"])
    cli.main()
    assert 'Could not find config option "web_spier_distance"' in caplog.text
    assert 'Did you mean "web.spider_distance"?' in caplog.text


def test_cli_module_validation(monkeypatch, caplog):
    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)

    # incorrect module
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-m", "dnsbrutes"])
    cli.main()
    assert 'Could not find scan module "dnsbrutes"' in caplog.text
    assert 'Did you mean "dnsbrute"?' in caplog.text

    # incorrect excluded module
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-em", "dnsbrutes"])
    cli.main()
    assert 'Could not find module "dnsbrutes"' in caplog.text
    assert 'Did you mean "dnsbrute"?' in caplog.text

    # incorrect output module
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-om", "neoo4j"])
    cli.main()
    assert 'Could not find output module "neoo4j"' in caplog.text
    assert 'Did you mean "neo4j"?' in caplog.text

    # output module setup failed
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-om", "websocket", "-c", "modules.websocket.url=", "-y"])
    cli.main()
    lines = caplog.text.splitlines()
    assert "Loaded 6/6 output modules, (csv,json,python,stdout,txt,websocket)" in caplog.text
    assert 1 == len(
        [
            l
            for l in lines
            if l.startswith("WARNING  bbot.scanner:scanner.py")
            and l.endswith("Setup hard-failed for websocket: Must set URL")
        ]
    )
    assert 1 == len(
        [
            l
            for l in lines
            if l.startswith("WARNING  bbot.modules.output.websocket:base.py") and l.endswith("Setting error state")
        ]
    )
    assert 1 == len(
        [
            l
            for l in lines
            if l.startswith("ERROR    bbot.cli:cli.py")
            and l.endswith("Setup hard-failed for 1 modules (websocket) (--force to run module anyway)")
        ]
    )

    # only output module setup failed
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr(
        "sys.argv",
        ["bbot", "-om", "websocket", "-em", "python,stdout,csv,json,txt", "-c", "modules.websocket.url=", "-y"],
    )
    cli.main()
    lines = caplog.text.splitlines()
    assert "Loaded 1/1 output modules, (websocket)" in caplog.text
    assert 1 == len(
        [
            l
            for l in lines
            if l.startswith("WARNING  bbot.scanner:scanner.py")
            and l.endswith("Setup hard-failed for websocket: Must set URL")
        ]
    )
    assert 1 == len(
        [
            l
            for l in lines
            if l.startswith("WARNING  bbot.modules.output.websocket:base.py") and l.endswith("Setting error state")
        ]
    )
    assert 1 == len(
        [
            l
            for l in lines
            if l.startswith("ERROR    bbot.cli:cli.py") and l.endswith("Failed to load output modules. Aborting.")
        ]
    )

    # incorrect flag
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-f", "subdomainenum"])
    cli.main()
    assert 'Could not find flag "subdomainenum"' in caplog.text
    assert 'Did you mean "subdomain-enum"?' in caplog.text

    # incorrect excluded flag
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-ef", "subdomainenum"])
    cli.main()
    assert 'Could not find flag "subdomainenum"' in caplog.text
    assert 'Did you mean "subdomain-enum"?' in caplog.text

    # incorrect required flag
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-rf", "subdomainenum"])
    cli.main()
    assert 'Could not find flag "subdomainenum"' in caplog.text
    assert 'Did you mean "subdomain-enum"?' in caplog.text


def test_cli_presets(monkeypatch, capsys, caplog):
    import yaml

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)

    # show current preset
    monkeypatch.setattr("sys.argv", ["bbot", "-c", "web.http_proxy=currentpresettest", "--current-preset"])
    cli.main()
    captured = capsys.readouterr()
    assert "    http_proxy: currentpresettest" in captured.out

    # show current preset (full)
    monkeypatch.setattr("sys.argv", ["bbot", "-c" "modules.c99.api_key=asdf", "--current-preset-full"])
    cli.main()
    captured = capsys.readouterr()
    assert "      api_key: asdf" in captured.out

    preset_dir = bbot_test_dir / "test_cli_presets"
    preset_dir.mkdir(exist_ok=True)

    preset1_file = preset_dir / "cli_preset1.conf"
    with open(preset1_file, "w") as f:
        f.write(
            """
config:
  web:
    http_proxy: http://proxy1
        """
        )

    preset2_file = preset_dir / "cli_preset2.yml"
    with open(preset2_file, "w") as f:
        f.write(
            """
config:
  web:
    http_proxy: http://proxy2
        """
        )

    # test reading single preset
    monkeypatch.setattr("sys.argv", ["bbot", "-p", str(preset1_file.resolve()), "--current-preset"])
    cli.main()
    captured = capsys.readouterr()
    stdout_preset = yaml.safe_load(captured.out)
    assert stdout_preset["config"]["web"]["http_proxy"] == "http://proxy1"

    # preset overrides preset
    monkeypatch.setattr(
        "sys.argv", ["bbot", "-p", str(preset2_file.resolve()), str(preset1_file.resolve()), "--current-preset"]
    )
    cli.main()
    captured = capsys.readouterr()
    stdout_preset = yaml.safe_load(captured.out)
    assert stdout_preset["config"]["web"]["http_proxy"] == "http://proxy1"

    # override other way
    monkeypatch.setattr(
        "sys.argv", ["bbot", "-p", str(preset1_file.resolve()), str(preset2_file.resolve()), "--current-preset"]
    )
    cli.main()
    captured = capsys.readouterr()
    stdout_preset = yaml.safe_load(captured.out)
    assert stdout_preset["config"]["web"]["http_proxy"] == "http://proxy2"

    # cli config overrides all presets
    monkeypatch.setattr(
        "sys.argv",
        [
            "bbot",
            "-p",
            str(preset1_file.resolve()),
            str(preset2_file.resolve()),
            "-c",
            "web.http_proxy=asdf",
            "--current-preset",
        ],
    )
    cli.main()
    captured = capsys.readouterr()
    stdout_preset = yaml.safe_load(captured.out)
    assert stdout_preset["config"]["web"]["http_proxy"] == "asdf"

    # invalid preset
    caplog.clear()
    assert not caplog.text
    monkeypatch.setattr("sys.argv", ["bbot", "-p", "asdfasdfasdf", "-y"])
    cli.main()
    assert "file does not exist. Use -lp to list available presets" in caplog.text

    preset1_file.unlink()
    preset2_file.unlink()

    # test output dir preset
    output_dir_preset_file = bbot_test_dir / "output_dir_preset.yml"
    scan_name = "cli_output_dir_test"
    output_dir = bbot_test_dir / "cli_output_dir_preset"
    scan_dir = output_dir / scan_name
    output_file = scan_dir / "output.txt"

    with open(output_dir_preset_file, "w") as f:
        f.write(
            f"""
output_dir: {output_dir}
scan_name: {scan_name}
        """
        )

    assert not output_dir.exists()
    assert not scan_dir.exists()
    assert not output_file.exists()

    monkeypatch.setattr("sys.argv", ["bbot", "-p", str(output_dir_preset_file.resolve()), "--current-preset"])
    cli.main()
    captured = capsys.readouterr()
    stdout_preset = yaml.safe_load(captured.out)
    assert stdout_preset["output_dir"] == str(output_dir)
    assert stdout_preset["scan_name"] == scan_name

    shutil.rmtree(output_dir, ignore_errors=True)
    shutil.rmtree(scan_dir, ignore_errors=True)
    shutil.rmtree(output_file, ignore_errors=True)

    assert not output_dir.exists()
    assert not scan_dir.exists()
    assert not output_file.exists()

    monkeypatch.setattr("sys.argv", ["bbot", "-p", str(output_dir_preset_file.resolve())])
    cli.main()
    captured = capsys.readouterr()
    assert output_dir.is_dir()
    assert scan_dir.is_dir()
    assert output_file.is_file()

    shutil.rmtree(output_dir, ignore_errors=True)
    shutil.rmtree(scan_dir, ignore_errors=True)
    shutil.rmtree(output_file, ignore_errors=True)
    output_dir_preset_file.unlink()



from bbot.modules.templates.subdomain_enum import subdomain_enum


class subdomaincenter(subdomain_enum):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {
        "description": "Query subdomain.center's API for subdomains",
        "created_date": "2023-07-26",
        "author": "@TheTechromancer",
    }

    base_url = "https://api.subdomain.center"
    retries = 2

    async def sleep(self, time_to_wait):
        self.info(f"Sleeping for {time_to_wait} seconds to avoid rate limit")
        await self.helpers.sleep(time_to_wait)

    async def request_url(self, query):
        url = f"{self.base_url}/?domain={self.helpers.quote(query)}"
        response = None
        status_code = 0
        for i, _ in enumerate(range(self.retries + 1)):
            if i > 0:
                self.verbose(f"Retry #{i} for {query} after response code {status_code}")
            response = await self.helpers.request(url, timeout=self.http_timeout + 30)
            status_code = getattr(response, "status_code", 0)
            if status_code == 429:
                await self.sleep(20)
            else:
                break
        return response

    def parse_results(self, r, query):
        results = set()
        json = r.json()
        if json and isinstance(json, list):
            results = set(json)
        return results



from .base import ModuleTestBase


class TestNeo4j(ModuleTestBase):
    config_overrides = {"modules": {"neo4j": {"uri": "bolt://127.0.0.1:11111"}}}

    async def setup_before_prep(self, module_test):
        # install neo4j
        deps_pip = module_test.preloaded["neo4j"]["deps"]["pip"]
        await module_test.scan.helpers.depsinstaller.pip_install(deps_pip)

        self.neo4j_used = False

        class MockResult:
            async def data(s):
                self.neo4j_used = True
                return [
                    {
                        "neo4j_id": "4:ee79a477-5f5b-445a-9def-7c051b2a533c:115",
                        "event_id": "DNS_NAME:c8fab50640cb87f8712d1998ecc78caf92b90f71",
                    }
                ]

        class MockSession:
            async def run(s, *args, **kwargs):
                return MockResult()

            async def close(self):
                pass

        class MockDriver:
            def __init__(self, *args, **kwargs):
                pass

            def session(self, *args, **kwargs):
                return MockSession()

            async def close(self):
                pass

        module_test.monkeypatch.setattr("neo4j.AsyncGraphDatabase.driver", MockDriver)

    def check(self, module_test, events):
        assert self.neo4j_used == True



from bbot.modules.base import BaseModule


class emailformat(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["EMAIL_ADDRESS"]
    flags = ["passive", "email-enum", "safe"]
    meta = {
        "description": "Query email-format.com for email addresses",
        "created_date": "2022-07-11",
        "author": "@TheTechromancer",
    }
    in_scope_only = False
    per_domain_only = True

    base_url = "https://www.email-format.com"

    async def handle_event(self, event):
        _, query = self.helpers.split_domain(event.data)
        url = f"{self.base_url}/d/{self.helpers.quote(query)}/"
        r = await self.request_with_fail_count(url)
        if not r:
            return
        for email in await self.helpers.re.extract_emails(r.text):
            if email.endswith(query):
                await self.emit_event(
                    email,
                    "EMAIL_ADDRESS",
                    parent=event,
                    context=f'{{module}} searched email-format.com for "{query}" and found {{event.type}}: {{event.data}}',
                )



from .base import ModuleTestBase, tempwordlist


class TestDnsbrute(ModuleTestBase):
    subdomain_wordlist = tempwordlist(["www", "asdf"])
    blacklist = ["api.asdf.blacklanternsecurity.com"]
    config_overrides = {"modules": {"dnsbrute": {"wordlist": str(subdomain_wordlist), "max_depth": 3}}}

    async def setup_after_prep(self, module_test):

        old_run_live = module_test.scan.helpers.run_live

        async def new_run_live(*command, check=False, text=True, **kwargs):
            if "massdns" in command[:2]:
                _input = [l async for l in kwargs["input"]]
                if "asdf.blacklanternsecurity.com" in _input:
                    yield """{"name": "asdf.blacklanternsecurity.com.", "type": "A", "class": "IN", "status": "NOERROR", "rx_ts": 1713974911725326170, "data": {"answers": [{"ttl": 86400, "type": "A", "class": "IN", "name": "asdf.blacklanternsecurity.com.", "data": "1.2.3.4."}]}, "flags": ["rd", "ra"], "resolver": "195.226.187.130:53", "proto": "UDP"}"""
            else:
                async for _ in old_run_live(*command, check=False, text=True, **kwargs):
                    yield _

        module_test.monkeypatch.setattr(module_test.scan.helpers, "run_live", new_run_live)

        await module_test.mock_dns(
            {
                "blacklanternsecurity.com": {"A": ["4.3.2.1"]},
                "asdf.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
            }
        )

        module = module_test.module
        scan = module_test.scan

        # test query logic
        event = scan.make_event("blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "blacklanternsecurity.com"
        event = scan.make_event("asdf.blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "blacklanternsecurity.com"
        event = scan.make_event("api.asdf.blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "asdf.blacklanternsecurity.com"
        event = scan.make_event("test.api.asdf.blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "asdf.blacklanternsecurity.com"

        assert module.dedup_strategy == "lowest_parent"
        module.dedup_strategy = "highest_parent"
        event = scan.make_event("blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "blacklanternsecurity.com"
        event = scan.make_event("asdf.blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "blacklanternsecurity.com"
        event = scan.make_event("api.asdf.blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "blacklanternsecurity.com"
        event = scan.make_event("test.api.asdf.blacklanternsecurity.com", "DNS_NAME", dummy=True)
        assert module.make_query(event) == "blacklanternsecurity.com"
        module.dedup_strategy = "lowest_parent"

        # test recursive brute-force event filtering
        event = module_test.scan.make_event("blacklanternsecurity.com", "DNS_NAME", parent=module_test.scan.root_event)
        event.scope_distance = 0
        result, reason = await module_test.module.filter_event(event)
        assert result == True
        event = module_test.scan.make_event(
            "www.blacklanternsecurity.com", "DNS_NAME", parent=module_test.scan.root_event
        )
        event.scope_distance = 0
        result, reason = await module_test.module.filter_event(event)
        assert result == True
        event = module_test.scan.make_event(
            "test.www.blacklanternsecurity.com", "DNS_NAME", parent=module_test.scan.root_event
        )
        event.scope_distance = 0
        result, reason = await module_test.module.filter_event(event)
        assert result == True
        event = module_test.scan.make_event(
            "asdf.test.www.blacklanternsecurity.com", "DNS_NAME", parent=module_test.scan.root_event
        )
        event.scope_distance = 0
        result, reason = await module_test.module.filter_event(event)
        assert result == True
        event = module_test.scan.make_event(
            "wat.asdf.test.www.blacklanternsecurity.com", "DNS_NAME", parent=module_test.scan.root_event
        )
        event.scope_distance = 0
        result, reason = await module_test.module.filter_event(event)
        assert result == False
        assert reason == f"subdomain depth of *.asdf.test.www.blacklanternsecurity.com (4) > max_depth (3)"
        event = module_test.scan.make_event(
            "hmmm.ptr1234.blacklanternsecurity.com", "DNS_NAME", parent=module_test.scan.root_event
        )
        event.scope_distance = 0
        result, reason = await module_test.module.filter_event(event)
        assert result == False
        assert reason == f'"ptr1234.blacklanternsecurity.com" looks like an autogenerated PTR'

    def check(self, module_test, events):
        assert len(events) == 4
        assert 1 == len(
            [e for e in events if e.data == "asdf.blacklanternsecurity.com" and str(e.module) == "dnsbrute"]
        )



# based on https://github.com/ElSicarius/interactsh-python/blob/main/sources/interactsh.py
import json
import base64
import random
import asyncio
import logging
import traceback
from uuid import uuid4

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Cipher import AES, PKCS1_OAEP

from bbot.errors import InteractshError

log = logging.getLogger("bbot.core.helpers.interactsh")

server_list = ["oast.pro", "oast.live", "oast.site", "oast.online", "oast.fun", "oast.me"]


class Interactsh:
    """
    A pure python implementation of ProjectDiscovery's interact.sh.

    *"Interactsh is an open-source tool for detecting out-of-band interactions. It is a tool designed to detect vulnerabilities that cause external interactions."*

    - https://app.interactsh.com
    - https://github.com/projectdiscovery/interactsh

    This class facilitates interactions with the interact.sh service for
    out-of-band data exfiltration and vulnerability confirmation. It allows
    for customization by accepting server and token parameters from the
    configuration provided by `parent_helper`.

    Attributes:
        parent_helper (ConfigAwareHelper): An instance of a helper class containing configuration data.
        server (str): The server to be used. If None (the default), a random server will be chosen from a predetermined list.
        correlation_id (str): An identifier to correlate requests and responses. Default is None.
        custom_server (str): Optional. A custom interact.sh server. Loaded from configuration.
        token (str): Optional. A token for interact.sh API. Loaded from configuration.
        _poll_task (AsyncTask): The task responsible for polling the interact.sh server.

    Examples:
        ```python
        # instantiate interact.sh client (no requests are sent yet)
        >>> interactsh_client = self.helpers.interactsh()
        # register with an interact.sh server
        >>> interactsh_domain = await interactsh_client.register()
        [INFO] Registering with interact.sh server: oast.me
        [INFO] Successfully registered to interactsh server oast.me with correlation_id rg99x2f860h5466ou3so [rg99x2f860h5466ou3so86i07n1m3013k.oast.me]
        # simulate an out-of-band interaction
        >>> await self.helpers.request(f"https://{interactsh_domain}/test")
        # wait for out-of-band interaction to be registered
        >>> await asyncio.sleep(10)
        >>> data_list = await interactsh_client.poll()
        >>> print(data_list)
        [
            {
                "protocol": "dns",
                "unique-id": "rg99x2f860h5466ou3so86i07n1m3013k",
                "full-id": "rg99x2f860h5466ou3so86i07n1m3013k",
                "q-type": "A",
                "raw-request": "...",
                "remote-address": "1.2.3.4",
                "timestamp": "2023-09-15T21:09:23.187226851Z"
            },
            {
                "protocol": "http",
                "unique-id": "rg99x2f860h5466ou3so86i07n1m3013k",
                "full-id": "rg99x2f860h5466ou3so86i07n1m3013k",
                "raw-request": "GET /test HTTP/1.1 ...",
                "remote-address": "1.2.3.4",
                "timestamp": "2023-09-15T21:09:24.155677967Z"
            }
        ]
        # finally, shut down the client
        >>> await interactsh_client.deregister()
        ```
    """

    def __init__(self, parent_helper, poll_interval=10):
        self.parent_helper = parent_helper
        self.server = None
        self.correlation_id = None
        self.custom_server = self.parent_helper.config.get("interactsh_server", None)
        self.token = self.parent_helper.config.get("interactsh_token", None)
        self.poll_interval = poll_interval
        self._poll_task = None

    async def register(self, callback=None):
        """
        Registers the instance with an interact.sh server and sets up polling.

        Generates RSA keys for secure communication, builds a correlation ID,
        and sends a POST request to an interact.sh server to register. Optionally,
        starts an asynchronous polling task to listen for interactions.

        Args:
            callback (callable, optional): A function to be called each time new interactions are received.

        Returns:
            str: The registered domain for out-of-band interactions.

        Raises:
            InteractshError: If registration with an interact.sh server fails.

        Examples:
            >>> interactsh_client = self.helpers.interactsh()
            >>> registered_domain = await interactsh_client.register()
            [INFO] Registering with interact.sh server: oast.me
            [INFO] Successfully registered to interactsh server oast.me with correlation_id rg99x2f860h5466ou3so [rg99x2f860h5466ou3so86i07n1m3013k.oast.me]
        """
        rsa = RSA.generate(1024)

        self.public_key = rsa.publickey().exportKey()
        self.private_key = rsa.exportKey()

        encoded_public_key = base64.b64encode(self.public_key).decode("utf8")

        uuid = uuid4().hex.ljust(33, "a")
        guid = "".join(i if i.isdigit() else chr(ord(i) + random.randint(0, 20)) for i in uuid)

        self.correlation_id = guid[:20]
        self.secret = str(uuid4())
        headers = {}

        if self.custom_server:
            if not self.token:
                log.verbose("Interact.sh token is not set")
            else:
                headers["Authorization"] = self.token
            self.server_list = [str(self.custom_server)]
        else:
            self.server_list = random.sample(server_list, k=len(server_list))
        for server in self.server_list:
            log.info(f"Registering with interact.sh server: {server}")
            data = {
                "public-key": encoded_public_key,
                "secret-key": self.secret,
                "correlation-id": self.correlation_id,
            }
            r = await self.parent_helper.request(
                f"https://{server}/register", headers=headers, json=data, method="POST"
            )
            if r is None:
                continue
            try:
                msg = r.json().get("message", "")
                assert "registration successful" in msg
            except Exception:
                log.debug(f"Failed to register with interactsh server {self.server}")
                continue
            self.server = server
            self.domain = f"{guid}.{self.server}"
            break

        if not self.server:
            raise InteractshError(f"Failed to register with an interactsh server")

        log.info(
            f"Successfully registered to interactsh server {self.server} with correlation_id {self.correlation_id} [{self.domain}]"
        )

        if callable(callback):
            self._poll_task = asyncio.create_task(self.poll_loop(callback))

        return self.domain

    async def deregister(self):
        """
        Deregisters the instance from the interact.sh server and cancels the polling task.

        Sends a POST request to the server to deregister, using the correlation ID
        and secret key generated during registration. Optionally, if a polling
        task was started, it is cancelled.

        Raises:
            InteractshError: If required information is missing or if deregistration fails.

        Examples:
            >>> await interactsh_client.deregister()
        """
        if not self.server or not self.correlation_id or not self.secret:
            raise InteractshError(f"Missing required information to deregister")

        headers = {}
        if self.token:
            headers["Authorization"] = self.token

        data = {"secret-key": self.secret, "correlation-id": self.correlation_id}

        r = await self.parent_helper.request(
            f"https://{self.server}/deregister", headers=headers, json=data, method="POST"
        )

        if self._poll_task is not None:
            self._poll_task.cancel()

        if "success" not in getattr(r, "text", ""):
            raise InteractshError(f"Failed to de-register with interactsh server {self.server}")

    async def poll(self):
        """
        Polls the interact.sh server for interactions tied to the current instance.

        Sends a GET request to the server to fetch interactions associated with the
        current correlation_id and secret key. Returned interactions are decrypted
        using an AES key provided by the server response.

        Raises:
            InteractshError: If required information for polling is missing.

        Returns:
            list: A list of decrypted interaction data dictionaries.

        Examples:
            >>> data_list = await interactsh_client.poll()
            >>> print(data_list)
            [
                {
                    "protocol": "dns",
                    "unique-id": "rg99x2f860h5466ou3so86i07n1m3013k",
                    ...
                },
                ...
            ]
        """
        if not self.server or not self.correlation_id or not self.secret:
            raise InteractshError(f"Missing required information to poll")

        headers = {}
        if self.token:
            headers["Authorization"] = self.token

        try:
            r = await self.parent_helper.request(
                f"https://{self.server}/poll?id={self.correlation_id}&secret={self.secret}", headers=headers
            )
            if r is None:
                raise InteractshError("Error polling interact.sh: No response from server")

            ret = []
            data_list = r.json().get("data", None)
            if data_list:
                aes_key = r.json()["aes_key"]

                for data in data_list:
                    decrypted_data = self._decrypt(aes_key, data)
                    ret.append(decrypted_data)
            return ret
        except Exception as e:
            raise InteractshError(f"Error polling interact.sh: {e}")

    async def poll_loop(self, callback):
        """
        Starts a polling loop to continuously check for interactions with the interact.sh server.

        Continuously polls the interact.sh server for interactions tied to the current instance,
        using the `poll` method. When interactions are received, it executes the given callback
        function with each interaction data.

        Parameters:
            callback (callable): The function to be called for every interaction received from the server.

        Returns:
            awaitable: An awaitable object that executes the internal `_poll_loop` method.

        Examples:
            >>> await interactsh_client.poll_loop(my_callback)
        """
        async with self.parent_helper.scan._acatch(context=self._poll_loop):
            return await self._poll_loop(callback)

    async def _poll_loop(self, callback):
        while 1:
            if self.parent_helper.scan.stopping:
                await asyncio.sleep(1)
                continue
            data_list = []
            try:
                data_list = await self.poll()
            except InteractshError as e:
                log.warning(e)
                log.trace(traceback.format_exc())
            if not data_list:
                await asyncio.sleep(self.poll_interval)
                continue
            for data in data_list:
                if data:
                    await self.parent_helper.execute_sync_or_async(callback, data)

    def _decrypt(self, aes_key, data):
        """
        Decrypts and returns the data received from the interact.sh server.

        Uses RSA and AES for decrypting the data. RSA with PKCS1_OAEP and SHA256 is used to decrypt the AES key,
        and then AES (CFB mode) is used to decrypt the actual data payload.

        Parameters:
            aes_key (str): The AES key for decryption, encrypted with RSA and base64 encoded.
            data (str): The data payload to decrypt, which is base64 encoded and AES encrypted.

        Returns:
            dict: The decrypted data, loaded as a JSON object.

        Examples:
            >>> decrypted_data = self._decrypt(aes_key, data)
        """
        private_key = RSA.importKey(self.private_key)
        cipher = PKCS1_OAEP.new(private_key, hashAlgo=SHA256)
        aes_plain_key = cipher.decrypt(base64.b64decode(aes_key))
        decode = base64.b64decode(data)
        bs = AES.block_size
        iv = decode[:bs]
        cryptor = AES.new(key=aes_plain_key, mode=AES.MODE_CFB, IV=iv, segment_size=128)
        plain_text = cryptor.decrypt(decode)
        return json.loads(plain_text[16:])



import sys

from bbot.modules.base import BaseModule


"""
wrapper for https://github.com/defparam/smuggler.git
"""


class smuggler(BaseModule):
    watched_events = ["URL"]
    produced_events = ["FINDING"]
    flags = ["active", "aggressive", "slow", "web-thorough"]
    meta = {"description": "Check for HTTP smuggling", "created_date": "2022-07-06", "author": "@liquidsec"}

    in_scope_only = True
    per_hostport_only = True

    deps_ansible = [
        {
            "name": "Get smuggler repo",
            "git": {"repo": "https://github.com/defparam/smuggler.git", "dest": "#{BBOT_TOOLS}/smuggler"},
        }
    ]

    async def handle_event(self, event):
        command = [
            sys.executable,
            f"{self.scan.helpers.tools_dir}/smuggler/smuggler.py",
            "--no-color",
            "-q",
            "-u",
            event.data,
        ]
        async for line in self.run_process_live(command):
            for f in line.split("\r"):
                if "Issue Found" in f:
                    technique = f.split(":")[0].rstrip()
                    text = f.split(":")[1].split("-")[0].strip()
                    description = f"[HTTP SMUGGLER] [{text}] Technique: {technique}"
                    await self.emit_event(
                        {"host": str(event.host), "url": event.data, "description": description},
                        "FINDING",
                        parent=event,
                        context=f"{{module}} scanned {event.data} and found HTTP smuggling ({{event.type}}): {text}",
                    )



from bbot.modules.templates.subdomain_enum import subdomain_enum


class otx(subdomain_enum):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {
        "description": "Query otx.alienvault.com for subdomains",
        "created_date": "2022-08-24",
        "author": "@TheTechromancer",
    }

    base_url = "https://otx.alienvault.com"

    def request_url(self, query):
        url = f"{self.base_url}/api/v1/indicators/domain/{self.helpers.quote(query)}/passive_dns"
        return self.request_with_fail_count(url)

    def parse_results(self, r, query):
        j = r.json()
        if isinstance(j, dict):
            for entry in j.get("passive_dns", []):
                subdomain = entry.get("hostname", "")
                if subdomain:
                    yield subdomain



class BBOTError(Exception):
    pass


class ScanError(BBOTError):
    pass


class ValidationError(BBOTError):
    pass


class ConfigLoadError(BBOTError):
    pass


class HttpCompareError(BBOTError):
    pass


class DirectoryCreationError(BBOTError):
    pass


class DirectoryDeletionError(BBOTError):
    pass


class NTLMError(BBOTError):
    pass


class InteractshError(BBOTError):
    pass


class WordlistError(BBOTError):
    pass


class CurlError(BBOTError):
    pass


class PresetNotFoundError(BBOTError):
    pass


class EnableModuleError(BBOTError):
    pass


class EnableFlagError(BBOTError):
    pass


class BBOTArgumentError(BBOTError):
    pass


class PresetConditionError(BBOTError):
    pass


class PresetAbortError(PresetConditionError):
    pass


class BBOTEngineError(BBOTError):
    pass


class WebError(BBOTEngineError):
    pass


class DNSError(BBOTEngineError):
    pass


class ExcavateError(BBOTError):
    pass
